@@ -50,10 +50,11 @@
 #include <string.h>
 #include <dirent.h>
 #include <attr/xattr.h>
+#include <linux/xattr.h>
 #include <getopt.h>
 #include <keyutils.h>
-#include <asm/byteorder.h>
 #include <ctype.h>
+#include <termios.h>
 
 #include <openssl/sha.h>
 #include <openssl/pem.h>
@@ -65,14 +66,27 @@
 
 #include "imaevm.h"
 
-static char *evm_config_xattrnames[] = {
-	"security.selinux",
-	"security.SMACK64",
-	"security.ima",
-	"security.capability",
+static char *evm_default_xattrs[] = {
+	XATTR_NAME_SELINUX,
+	XATTR_NAME_SMACK,
+	XATTR_NAME_IMA,
+	XATTR_NAME_CAPS,
 	NULL
 };
 
+static char *evm_extra_smack_xattrs[] = {
+	XATTR_NAME_SELINUX,
+	XATTR_NAME_SMACK,
+	XATTR_NAME_SMACKEXEC,
+	XATTR_NAME_SMACKTRANSMUTE,
+	XATTR_NAME_SMACKMMAP,
+	XATTR_NAME_IMA,
+	XATTR_NAME_CAPS,
+	NULL
+};
+
+static char **evm_config_xattrnames = evm_default_xattrs;
+
 struct command {
 	char *name;
 	int (*func)(struct command *cmd);
@@ -84,17 +98,21 @@ struct command {
 static int g_argc;
 static char **g_argv;
 static int xattr = 1;
+static bool check_xattr;
 static int sigdump;
 static int digest;
 static int digsig;
-static char *keypass;
 static int sigfile;
-static int x509 = 1;
-static char *uuid_str = "+";
+static char *uuid_str;
 static char *search_type;
 static int recursive;
 static int msize;
 static dev_t fs_dev;
+static bool evm_immutable;
+
+#define HMAC_FLAG_UUID			0x0001
+#define HMAC_FLAG_UUID_SET		0x0002
+static unsigned long hmac_flags = HMAC_FLAG_UUID;
 
 typedef int (*find_cb_t)(const char *path);
 static int find(const char *path, int dts, find_cb_t func);
@@ -105,10 +123,6 @@ static int find(const char *path, int dts, find_cb_t func);
 #define CHR_MASK	(1 << DT_CHR)
 #define BLK_MASK	(1 << DT_BLK)
 
-typedef int (*sign_hash_fn_t)(const char *algo, const unsigned char *hash, int size, const char *keyfile, unsigned char *sig);
-
-static sign_hash_fn_t sign_hash;
-
 struct command cmds[];
 static void print_usage(struct command *cmd);
 
@@ -127,7 +141,7 @@ static int bin2file(const char *file, const char *ext, const unsigned char *data
 
 	fp = fopen(name, "w");
 	if (!fp) {
-		log_err("Unable to open %s for writing\n", name);
+		log_err("Failed to open: %s\n", name);
 		return -1;
 	}
 	err = fwrite(data, len, 1, fp);
@@ -152,7 +166,7 @@ static unsigned char *file2bin(const char *file, const char *ext, int *size)
 	len = get_filesize(name);
 	fp = fopen(name, "r");
 	if (!fp) {
-		log_err("Unable to open %s\n", name);
+		log_err("Failed to open: %s\n", name);
 		return NULL;
 	}
 	data = malloc(len);
@@ -164,215 +178,6 @@ static unsigned char *file2bin(const char *file, const char *ext, int *size)
 	return data;
 }
 
-/*
- * Create binary key representation suitable for kernel
- */
-static int key2bin(RSA *key, unsigned char *pub)
-{
-	int len, b, offset = 0;
-	struct pubkey_hdr *pkh = (struct pubkey_hdr *)pub;
-
-	/* add key header */
-	pkh->version = 1;
-	pkh->timestamp = 0;	/* PEM has no timestamp?? */
-	pkh->algo = PUBKEY_ALGO_RSA;
-	pkh->nmpi = 2;
-
-	offset += sizeof(*pkh);
-
-	len = BN_num_bytes(key->n);
-	b = BN_num_bits(key->n);
-	pub[offset++] = b >> 8;
-	pub[offset++] = b & 0xff;
-	BN_bn2bin(key->n, &pub[offset]);
-	offset += len;
-
-	len = BN_num_bytes(key->e);
-	b = BN_num_bits(key->e);
-	pub[offset++] = b >> 8;
-	pub[offset++] = b & 0xff;
-	BN_bn2bin(key->e, &pub[offset]);
-	offset += len;
-
-	return offset;
-}
-
-static void calc_keyid_v1(uint8_t *keyid, char *str, const unsigned char *pkey, int len)
-{
-	uint8_t sha1[SHA_DIGEST_LENGTH];
-	uint64_t id;
-
-	SHA1(pkey, len, sha1);
-
-	/* sha1[12 - 19] is exactly keyid from gpg file */
-	memcpy(keyid, sha1 + 12, 8);
-	log_debug("keyid: ");
-	log_debug_dump(keyid, 8);
-
-	id = __be64_to_cpup((__be64 *) keyid);
-	sprintf(str, "%llX", (unsigned long long)id);
-	log_info("keyid: %s\n", str);
-}
-
-static void calc_keyid_v2(uint32_t *keyid, char *str, RSA *key)
-{
-	uint8_t sha1[SHA_DIGEST_LENGTH];
-	unsigned char *pkey = NULL;
-	int len;
-
-	len = i2d_RSAPublicKey(key, &pkey);
-
-	SHA1(pkey, len, sha1);
-
-	/* sha1[12 - 19] is exactly keyid from gpg file */
-	memcpy(keyid, sha1 + 16, 4);
-	log_debug("keyid: ");
-	log_debug_dump(keyid, 4);
-
-	sprintf(str, "%x", __be32_to_cpup(keyid));
-	log_info("keyid: %s\n", str);
-
-	free(pkey);
-}
-
-static RSA *read_priv_key(const char *keyfile)
-{
-	FILE *fp;
-	RSA *key;
-
-	fp = fopen(keyfile, "r");
-	if (!fp) {
-		log_err("Unable to open keyfile %s\n", keyfile);
-		return NULL;
-	}
-	key = PEM_read_RSAPrivateKey(fp, NULL, NULL, keypass);
-	if (!key)
-		log_err("PEM_read_RSAPrivateKey() failed\n");
-
-	fclose(fp);
-	return key;
-}
-
-static int get_hash_algo_v1(const char *algo)
-{
-
-	if (!strcmp(algo, "sha1"))
-		return DIGEST_ALGO_SHA1;
-	else if (!strcmp(algo, "sha256"))
-		return DIGEST_ALGO_SHA256;
-
-	return -1;
-}
-
-static int sign_hash_v1(const char *hashalgo, const unsigned char *hash, int size, const char *keyfile, unsigned char *sig)
-{
-	int len = -1, hashalgo_idx;
-	SHA_CTX ctx;
-	unsigned char pub[1024];
-	RSA *key;
-	char name[20];
-	unsigned char sighash[20];
-	struct signature_hdr *hdr = (struct signature_hdr *)sig;
-	uint16_t *blen;
-
-	log_info("hash: ");
-	log_dump(hash, size);
-
-	key = read_priv_key(keyfile);
-	if (!key)
-		return -1;
-
-	/* now create a new hash */
-	hdr->version = (uint8_t) DIGSIG_VERSION_1;
-	hdr->timestamp = time(NULL);
-	hdr->algo = PUBKEY_ALGO_RSA;
-	hashalgo_idx = get_hash_algo_v1(hashalgo);
-	if (hashalgo_idx < 0) {
-		log_err("Signature version 1 does not support hash algo %s\n",
-			hashalgo);
-		goto out;
-	}
-	hdr->hash = (uint8_t) hashalgo_idx;
-
-	len = key2bin(key, pub);
-	calc_keyid_v1(hdr->keyid, name, pub, len);
-
-	hdr->nmpi = 1;
-
-	SHA1_Init(&ctx);
-	SHA1_Update(&ctx, hash, size);
-	SHA1_Update(&ctx, hdr, sizeof(*hdr));
-	SHA1_Final(sighash, &ctx);
-	log_info("sighash: ");
-	log_dump(sighash, sizeof(sighash));
-
-	len = RSA_private_encrypt(sizeof(sighash), sighash, sig + sizeof(*hdr) + 2, key, RSA_PKCS1_PADDING);
-	if (len < 0) {
-		log_err("RSA_private_encrypt() failed: %d\n", len);
-		goto out;
-	}
-
-	/* we add bit length of the signature to make it gnupg compatible */
-	blen = (uint16_t *) (sig + sizeof(*hdr));
-	*blen = __cpu_to_be16(len << 3);
-	len += sizeof(*hdr) + 2;
-	log_info("evm/ima signature: %d bytes\n", len);
-	if (sigdump || params.verbose >= LOG_INFO)
-		dump(sig, len);
-out:
-	RSA_free(key);
-	return len;
-}
-
-static int sign_hash_v2(const char *algo, const unsigned char *hash, int size, const char *keyfile, unsigned char *sig)
-{
-	struct signature_v2_hdr *hdr = (struct signature_v2_hdr *)sig;
-	int len = -1;
-	RSA *key;
-	char name[20];
-	unsigned char *buf;
-	const struct RSA_ASN1_template *asn1;
-
-	log_info("hash: ");
-	log_dump(hash, size);
-
-	key = read_priv_key(keyfile);
-	if (!key)
-		return -1;
-
-	hdr->version = (uint8_t) DIGSIG_VERSION_2;
-	hdr->hash_algo = get_hash_algo(algo);
-
-	calc_keyid_v2(&hdr->keyid, name, key);
-
-	asn1 = &RSA_ASN1_templates[hdr->hash_algo];
-
-	buf = malloc(size + asn1->size);
-	if (!buf)
-		goto out;
-
-	memcpy(buf, asn1->data, asn1->size);
-	memcpy(buf + asn1->size, hash, size);
-	len = RSA_private_encrypt(size + asn1->size, buf, hdr->sig,
-				  key, RSA_PKCS1_PADDING);
-	if (len < 0) {
-		log_err("RSA_private_encrypt() failed: %d\n", len);
-		goto out;
-	}
-
-	/* we add bit length of the signature to make it gnupg compatible */
-	hdr->sig_size = __cpu_to_be16(len);
-	len += sizeof(*hdr);
-	log_info("evm/ima signature: %d bytes\n", len);
-	if (sigdump || params.verbose >= LOG_INFO)
-		dump(sig, len);
-out:
-	if (buf)
-		free(buf);
-	RSA_free(key);
-	return len;
-}
-
 static int find_xattr(const char *list, int list_size, const char *xattr)
 {
 	int len;
@@ -470,7 +275,7 @@ static int get_uuid(struct stat *st, char *uuid)
 	FILE *fp;
 	size_t len;
 
-	if (uuid_str[0] != '+')
+	if (hmac_flags & HMAC_FLAG_UUID_SET)
 		return pack_uuid(uuid_str, uuid);
 
 	dev = st->st_dev;
@@ -481,19 +286,18 @@ static int get_uuid(struct stat *st, char *uuid)
 	sprintf(path, "blkid -s UUID -o value /dev/block/%u:%u", major, minor);
 
 	fp = popen(path, "r");
-	if (!fp) {
-		log_err("popen() failed\n");
-		return -1;
-	}
+	if (!fp)
+		goto err;
 
 	len = fread(_uuid, 1, sizeof(_uuid), fp);
 	pclose(fp);
-	if (len != sizeof(_uuid)) {
-		log_err("fread() failed\n");
-		return -1;
-	}
+	if (len != sizeof(_uuid))
+		goto err;
 
 	return pack_uuid(_uuid, uuid);
+err:
+	log_err("Failed to read UUID. Root access might require.\n");
+	return -1;
 }
 
 static int calc_evm_hash(const char *file, unsigned char *hash)
@@ -512,26 +316,28 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 	int hmac_size;
 
 	if (lstat(file, &st)) {
-		log_err("lstat() failed\n");
+		log_err("Failed to stat: %s\n", file);
 		return -1;
 	}
 
-	if (S_ISREG(st.st_mode) || S_ISDIR(st.st_mode)) {
-		/* we cannot at the momement to get generation of special files..
-		 * kernel API does not support it */
-		int fd = open(file, 0);
-		if (fd < 0) {
-			log_err("Unable to open %s\n", file);
-			return -1;
-		}
-		if (ioctl(fd, EXT34_IOC_GETVERSION, &generation)) {
-			log_err("ioctl() failed\n");
-			return -1;
-		}
-		close(fd);
-	}
+	if (!evm_immutable) {
+		if (S_ISREG(st.st_mode) || S_ISDIR(st.st_mode)) {
+			/* we cannot at the momement to get generation of special files..
+			 * kernel API does not support it */
+			int fd = open(file, 0);
 
-	log_info("generation: %u\n", generation);
+			if (fd < 0) {
+				log_err("Failed to open: %s\n", file);
+				return -1;
+			}
+			if (ioctl(fd, FS_IOC_GETVERSION, &generation)) {
+				log_err("ioctl() failed\n");
+				return -1;
+			}
+			close(fd);
+		}
+		log_info("generation: %u\n", generation);
+	}
 
 	list_size = llistxattr(file, list, sizeof(list));
 	if (list_size < 0) {
@@ -567,8 +373,16 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 
 	memset(&hmac_misc, 0, sizeof(hmac_misc));
 
-	if (msize == 0) {
+	if (evm_immutable) {
+		struct h_misc_digsig *hmac = (struct h_misc_digsig *)&hmac_misc;
+
+		hmac_size = sizeof(*hmac);
+		hmac->uid = st.st_uid;
+		hmac->gid = st.st_gid;
+		hmac->mode = st.st_mode;
+	} else if (msize == 0) {
 		struct h_misc *hmac = (struct h_misc *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -577,6 +391,7 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 		hmac->mode = st.st_mode;
 	} else if (msize == 64) {
 		struct h_misc_64 *hmac = (struct h_misc_64 *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -585,6 +400,7 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 		hmac->mode = st.st_mode;
 	} else {
 		struct h_misc_32 *hmac = (struct h_misc_32 *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -602,7 +418,7 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 		return 1;
 	}
 
-	if (*uuid_str != '-') {
+	if (!evm_immutable && (hmac_flags & HMAC_FLAG_UUID)) {
 		err = get_uuid(&st, uuid);
 		if (err)
 			return -1;
@@ -626,19 +442,29 @@ static int calc_evm_hash(const char *file, unsigned char *hash)
 static int sign_evm(const char *file, const char *key)
 {
 	unsigned char hash[20];
-	unsigned char sig[1024] = "\x03";
+	unsigned char sig[1024];
 	int len, err;
 
 	len = calc_evm_hash(file, hash);
 	if (len <= 1)
 		return len;
 
-	len = sign_hash("sha1", hash, len, key, sig + 1);
+	len = sign_hash("sha1", hash, len, key, NULL, sig + 1);
 	if (len <= 1)
 		return len;
 
+	/* add header */
+	len++;
+	sig[0] = EVM_IMA_XATTR_DIGSIG;
+
+	if (evm_immutable)
+		sig[1] = 3; /* immutable signature version */
+
+	if (sigdump || params.verbose >= LOG_INFO)
+		dump(sig, len);
+
 	if (xattr) {
-		err = lsetxattr(file, "security.evm", sig, len + 1, 0);
+		err = lsetxattr(file, "security.evm", sig, len, 0);
 		if (err < 0) {
 			log_err("setxattr failed: %s\n", file);
 			return err;
@@ -650,21 +476,33 @@ static int sign_evm(const char *file, const char *key)
 
 static int hash_ima(const char *file)
 {
-	unsigned char hash[65] = "\x01"; /* MAX hash size + 1 */
-	int len, err;
+	unsigned char hash[66]; /* MAX hash size + 2 */
+	int len, err, offset;
+	int algo = get_hash_algo(params.hash_algo);
 
-	len = ima_calc_hash(file, hash + 1);
+	if (algo > PKEY_HASH_SHA1) {
+		hash[0] = IMA_XATTR_DIGEST_NG;
+		hash[1] = algo;
+		offset = 2;
+	} else {
+		hash[0] = IMA_XATTR_DIGEST;
+		offset = 1;
+	}
+
+	len = ima_calc_hash(file, hash + offset);
 	if (len <= 1)
 		return len;
 
+	len += offset;
+
 	if (params.verbose >= LOG_INFO)
 		log_info("hash: ");
 
 	if (sigdump || params.verbose >= LOG_INFO)
-		dump(hash, len + 1);
+		dump(hash, len);
 
 	if (xattr) {
-		err = lsetxattr(file, "security.ima", hash, len + 1, 0);
+		err = lsetxattr(file, "security.ima", hash, len, 0);
 		if (err < 0) {
 			log_err("setxattr failed: %s\n", file);
 			return err;
@@ -674,35 +512,26 @@ static int hash_ima(const char *file)
 	return 0;
 }
 
-static int cmd_hash_ima(struct command *cmd)
-{
-	char *file = g_argv[optind++];
-
-	if (!file) {
-		log_err("Parameters missing\n");
-		print_usage(cmd);
-		return -1;
-	}
-
-	return hash_ima(file);
-}
-
 static int sign_ima(const char *file, const char *key)
 {
 	unsigned char hash[64];
-	unsigned char sig[1024] = "\x03";
+	unsigned char sig[1024];
 	int len, err;
 
 	len = ima_calc_hash(file, hash);
 	if (len <= 1)
 		return len;
 
-	len = sign_hash(params.hash_algo, hash, len, key, sig + 1);
+	len = sign_hash(params.hash_algo, hash, len, key, NULL, sig + 1);
 	if (len <= 1)
 		return len;
 
 	/* add header */
 	len++;
+	sig[0] = EVM_IMA_XATTR_DIGSIG;
+
+	if (sigdump || params.verbose >= LOG_INFO)
+		dump(sig, len);
 
 	if (sigfile)
 		bin2file(file, "sig", sig, len);
@@ -731,11 +560,13 @@ static int get_file_type(const char *path, const char *search_type)
 			dts |= DIR_MASK; break;
 		case 's':
 			dts |= BLK_MASK | CHR_MASK | LNK_MASK; break;
+		case 'x':
+			check_xattr = true; break;
 		case 'm':
 			/* stay within the same filesystem*/
 			err = lstat(path, &st);
 			if (err < 0) {
-				log_err("stat() failed\n");
+				log_err("Failed to stat: %s\n", path);
 				return err;
 			}
 			fs_dev = st.st_dev; /* filesystem to start from */
@@ -746,9 +577,39 @@ static int get_file_type(const char *path, const char *search_type)
 	return dts;
 }
 
+static int do_cmd(struct command *cmd, find_cb_t func)
+{
+	char *path = g_argv[optind++];
+	int err, dts = REG_MASK; /* only regular files by default */
+
+	if (!path) {
+		log_err("Parameters missing\n");
+		print_usage(cmd);
+		return -1;
+	}
+
+	if (recursive) {
+		if (search_type) {
+			dts = get_file_type(path, search_type);
+			if (dts < 0)
+				return dts;
+		}
+		err = find(path, dts, func);
+	} else {
+		err = func(path);
+	}
+
+	return err;
+}
+
+static int cmd_hash_ima(struct command *cmd)
+{
+	return do_cmd(cmd, hash_ima);
+}
+
 static int sign_ima_file(const char *file)
 {
-	char *key;
+	const char *key;
 
 	key = params.keyfile ? : "/etc/keys/privkey_evm.pem";
 
@@ -757,32 +618,13 @@ static int sign_ima_file(const char *file)
 
 static int cmd_sign_ima(struct command *cmd)
 {
-	char *file = g_argv[optind++];
-	int err, dts = REG_MASK; /* only regular files by default */
-
-	if (!file) {
-		log_err("Parameters missing\n");
-		print_usage(cmd);
-		return -1;
-	}
-
-	if (recursive) {
-		if (search_type) {
-			dts = get_file_type(file, search_type);
-			if (dts < 0)
-				return dts;
-		}
-		err = find(file, dts, sign_ima_file);
-	} else {
-		err = sign_ima_file(file);
-	}
-
-	return err;
+	return do_cmd(cmd, sign_ima_file);
 }
 
 static int cmd_sign_hash(struct command *cmd)
 {
-	char *key, *token, *line = NULL;
+	const char *key;
+	char *token, *line = NULL;
 	int hashlen = 0;
 	size_t line_len;
 	ssize_t len;
@@ -804,7 +646,7 @@ static int cmd_sign_hash(struct command *cmd)
 
 		hex2bin(hash, line, hashlen);
 		siglen = sign_hash(params.hash_algo, hash, hashlen/2,
-				 key, sig + 1);
+				 key, NULL, sig + 1);
 		if (siglen <= 1)
 			return siglen;
 
@@ -825,7 +667,7 @@ static int cmd_sign_hash(struct command *cmd)
 
 static int sign_evm_path(const char *file)
 {
-	char *key;
+	const char *key;
 	int err;
 
 	key = params.keyfile ? : "/etc/keys/privkey_evm.pem";
@@ -847,27 +689,7 @@ static int sign_evm_path(const char *file)
 
 static int cmd_sign_evm(struct command *cmd)
 {
-	char *path = g_argv[optind++];
-	int err, dts = REG_MASK; /* only regular files by default */
-
-	if (!path) {
-		log_err("Parameters missing\n");
-		print_usage(cmd);
-		return -1;
-	}
-
-	if (recursive) {
-		if (search_type) {
-			dts = get_file_type(path, search_type);
-			if (dts < 0)
-				return dts;
-		}
-		err = find(path, dts, sign_evm_path);
-	} else {
-		err = sign_evm_path(path);
-	}
-
-	return err;
+	return do_cmd(cmd, sign_evm_path);
 }
 
 static int verify_evm(const char *file)
@@ -882,12 +704,12 @@ static int verify_evm(const char *file)
 
 	len = lgetxattr(file, "security.evm", sig, sizeof(sig));
 	if (len < 0) {
-		log_err("getxattr failed\n");
+		log_err("getxattr failed: %s\n", file);
 		return len;
 	}
 
 	if (sig[0] != 0x03) {
-		log_err("security.evm has not signature\n");
+		log_err("security.evm has no signature\n");
 		return -1;
 	}
 
@@ -915,14 +737,14 @@ static int verify_ima(const char *file)
 	if (xattr) {
 		len = lgetxattr(file, "security.ima", sig, sizeof(sig));
 		if (len < 0) {
-			log_err("getxattr failed\n");
+			log_err("getxattr failed: %s\n", file);
 			return len;
 		}
 	}
 
 	if (sigfile) {
-		void *tmp;
-		tmp = file2bin(file, "sig", &len);
+		void *tmp = file2bin(file, "sig", &len);
+
 		memcpy(sig, tmp, len);
 		free(tmp);
 	}
@@ -954,8 +776,8 @@ static int cmd_import(struct command *cmd)
 
 	inkey = g_argv[optind++];
 	if (!inkey) {
-		inkey = x509 ? "/etc/keys/x509_evm.der" :
-			       "/etc/keys/pubkey_evm.pem";
+		inkey = params.x509 ? "/etc/keys/x509_evm.der" :
+				      "/etc/keys/pubkey_evm.pem";
 	} else
 		ring = g_argv[optind++];
 
@@ -963,7 +785,11 @@ static int cmd_import(struct command *cmd)
 
 	if (ring) {
 		if (ring[0] != '@') {
-			id = atoi(ring);
+			int base = 10;
+
+			if (ring[0] == '0' && ring[1] == 'x')
+				base = 16;
+			id = strtoul(ring, NULL, base);
 		} else {
 			if (strcmp(ring, "@t") == 0)
 				id = -1;
@@ -980,11 +806,11 @@ static int cmd_import(struct command *cmd)
 		}
 	}
 
-	key = read_pub_key(inkey, x509);
+	key = read_pub_key(inkey, params.x509);
 	if (!key)
 		return 1;
 
-	if (x509) {
+	if (params.x509) {
 		pub = file2bin(inkey, NULL, &len);
 		if (!pub)
 			goto out;
@@ -996,7 +822,7 @@ static int cmd_import(struct command *cmd)
 
 	log_info("Importing public key %s from file %s into keyring %d\n", name, inkey, id);
 
-	id = add_key(x509 ? "asymmetric" : "user", x509 ? NULL : name, pub, len, id);
+	id = add_key(params.x509 ? "asymmetric" : "user", params.x509 ? NULL : name, pub, len, id);
 	if (id < 0) {
 		log_err("add_key failed\n");
 		err = id;
@@ -1004,7 +830,7 @@ static int cmd_import(struct command *cmd)
 		log_info("keyid: %d\n", id);
 		printf("%d\n", id);
 	}
-	if (x509)
+	if (params.x509)
 		free(pub);
 out:
 	RSA_free(key);
@@ -1032,12 +858,12 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 
 	key = file2bin(keyfile, NULL, &keylen);
 	if (!key) {
-		log_err("Unable to read a key: %s\n\n", keyfile);
+		log_err("Failed to read a key: %s\n", keyfile);
 		return -1;
 	}
 
 	if (keylen > sizeof(evmkey)) {
-		log_err("key is too long\n");
+		log_err("key is too long: %d\n", keylen);
 		goto out;
 	}
 
@@ -1046,7 +872,7 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 	memset(evmkey + keylen, 0, sizeof(evmkey) - keylen);
 
 	if (lstat(file, &st)) {
-		log_err("lstat() failed\n");
+		log_err("Failed to stat: %s\n", file);
 		goto out;
 	}
 
@@ -1054,11 +880,12 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 		/* we cannot at the momement to get generation of special files..
 		 * kernel API does not support it */
 		int fd = open(file, 0);
+
 		if (fd < 0) {
-			log_err("Unable to open %s\n", file);
+			log_err("Failed to open %s\n", file);
 			goto out;
 		}
-		if (ioctl(fd, EXT34_IOC_GETVERSION, &generation)) {
+		if (ioctl(fd, FS_IOC_GETVERSION, &generation)) {
 			log_err("ioctl() failed\n");
 			goto out;
 		}
@@ -1069,7 +896,7 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 
 	list_size = llistxattr(file, list, sizeof(list));
 	if (list_size <= 0) {
-		log_err("llistxattr() failed\n");
+		log_err("llistxattr() failed: %s\n", file);
 		goto out;
 	}
 
@@ -1103,6 +930,7 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 
 	if (msize == 0) {
 		struct h_misc *hmac = (struct h_misc *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -1111,6 +939,7 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 		hmac->mode = st.st_mode;
 	} else if (msize == 64) {
 		struct h_misc_64 *hmac = (struct h_misc_64 *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -1119,6 +948,7 @@ static int calc_evm_hmac(const char *file, const char *keyfile, unsigned char *h
 		hmac->mode = st.st_mode;
 	} else {
 		struct h_misc_32 *hmac = (struct h_misc_32 *)&hmac_misc;
+
 		hmac_size = sizeof(*hmac);
 		hmac->ino = st.st_ino;
 		hmac->generation = generation;
@@ -1148,7 +978,7 @@ out:
 static int hmac_evm(const char *file, const char *key)
 {
 	unsigned char hash[20];
-	unsigned char sig[1024] = "\x02";
+	unsigned char sig[1024];
 	int len, err;
 
 	len = calc_evm_hmac(file, key, hash);
@@ -1160,6 +990,7 @@ static int hmac_evm(const char *file, const char *key)
 	memcpy(sig + 1, hash, len);
 
 	if (xattr) {
+		sig[0] = EVM_XATTR_HMAC;
 		err = lsetxattr(file, "security.evm", sig, len + 1, 0);
 		if (err < 0) {
 			log_err("setxattr failed: %s\n", file);
@@ -1172,7 +1003,7 @@ static int hmac_evm(const char *file, const char *key)
 
 static int cmd_hmac_evm(struct command *cmd)
 {
-	char *key, *file = g_argv[optind++];
+	const char *key, *file = g_argv[optind++];
 	int err;
 
 	if (!file) {
@@ -1205,13 +1036,13 @@ static int ima_fix(const char *path)
 
 	log_info("%s\n", path);
 
-	if (xattr) {
+	if (check_xattr) {
 		/* re-measuring takes a time
 		 * in some cases we can skip labeling if xattrs exists
 		 */
 		size = llistxattr(path, list, sizeof(buf));
 		if (size < 0) {
-			log_errno("llistxattr() failed: %s\n", path);
+			log_errno("Failed to read xattrs (llistxattr): %s\n", path);
 			return -1;
 		}
 		for (; size > 0; len++, size -= len, list += len) {
@@ -1227,7 +1058,7 @@ static int ima_fix(const char *path)
 
 	fd = open(path, O_RDONLY);
 	if (fd < 0) {
-		log_errno("%s open failed", path);
+		log_errno("Failed to open file: %s", path);
 		return -1;
 	}
 
@@ -1244,8 +1075,9 @@ static int find(const char *path, int dts, find_cb_t func)
 	if (fs_dev) {
 		struct stat st;
 		int err = lstat(path, &st);
+
 		if (err < 0) {
-			log_err("stat() failed\n");
+			log_err("Failed to stat: %s\n", path);
 			return err;
 		}
 		if (st.st_dev != fs_dev)
@@ -1254,12 +1086,12 @@ static int find(const char *path, int dts, find_cb_t func)
 
 	dir = opendir(path);
 	if (!dir) {
-		log_err("Unable to open %s\n", path);
+		log_err("Failed to open directory %s\n", path);
 		return -1;
 	}
 
 	if (fchdir(dirfd(dir))) {
-		log_err("Unable to chdir %s\n", path);
+		log_err("Failed to chdir %s\n", path);
 		return -1;
 	}
 
@@ -1274,7 +1106,7 @@ static int find(const char *path, int dts, find_cb_t func)
 	}
 
 	if (chdir("..")) {
-		log_err("Unable to chdir %s\n", path);
+		log_err("Failed to chdir: %s\n", path);
 		return -1;
 	}
 
@@ -1288,53 +1120,22 @@ static int find(const char *path, int dts, find_cb_t func)
 
 static int cmd_ima_fix(struct command *cmd)
 {
-	char *path = g_argv[optind++];
-	int err, dts = REG_MASK; /* only regular files by default */
-	struct stat st;
+	return do_cmd(cmd, ima_fix);
+}
 
-	if (!path) {
-		log_err("Parameters missing\n");
-		print_usage(cmd);
-		return -1;
-	}
-
-	xattr = 0; /* do not check xattrs, fix everything */
-
-	if (search_type) {
-		int i;
-
-		dts = 0;
-		for (i = 0; search_type[i]; i++) {
-			switch (search_type[i]) {
-			case 'f':
-				dts |= REG_MASK; break;
-			case 'd':
-				dts |= DIR_MASK; break;
-			case 's':
-				dts |= BLK_MASK | CHR_MASK | LNK_MASK; break;
-			case 'x':
-				/* check xattrs */
-				xattr = 1; break;
-			case 'm':
-				/* stay within the same filesystem*/
-				err = lstat(path, &st);
-				if (err < 0) {
-					log_err("stat() failed\n");
-					return err;
-				}
-				fs_dev = st.st_dev; /* filesystem to start from */
-				break;
-			}
-		}
-	}
-
-	err = find(path, dts, ima_fix);
-	if (err)
-		return err;
+static int ima_clear(const char *path)
+{
+	log_info("%s\n", path);
+	lremovexattr(path, "security.ima");
+	lremovexattr(path, "security.evm");
 
 	return 0;
 }
 
+static int cmd_ima_clear(struct command *cmd)
+{
+	return do_cmd(cmd, ima_clear);
+}
 
 static char *pcrs = "/sys/class/misc/tpm0/device/pcrs";
 
@@ -1492,7 +1293,7 @@ static int ima_measurement(const char *file)
 	uint8_t pcr10[SHA_DIGEST_LENGTH];
 	struct template_entry entry = { .template = 0 };
 	FILE *fp;
-	int err;
+	int err = -1;
 
 	memset(fox, 0xff, SHA_DIGEST_LENGTH);
 
@@ -1501,23 +1302,23 @@ static int ima_measurement(const char *file)
 
 	fp = fopen(file, "rb");
 	if (!fp) {
-		log_err("Unable to open measurement file\n");
+		log_err("Failed to open measurement file: %s\n", file);
 		return -1;
 	}
 
-	while ((err = fread(&entry.header, sizeof(entry.header), 1, fp))) {
+	while (fread(&entry.header, sizeof(entry.header), 1, fp)) {
 		ima_extend_pcr(pcr, entry.header.digest, SHA_DIGEST_LENGTH);
 
 		if (!fread(entry.name, entry.header.name_len, 1, fp)) {
 			log_err("Unable to read template name\n");
-			return -1;
+			goto out;
 		}
 
 		entry.name[entry.header.name_len] = '\0';
 
 		if (!fread(&entry.template_len, sizeof(entry.template_len), 1, fp)) {
 			log_err("Unable to read template length\n");
-			return -1;
+			goto out;
 		}
 
 		if (entry.template_buf_len < entry.template_len) {
@@ -1528,7 +1329,7 @@ static int ima_measurement(const char *file)
 
 		if (!fread(entry.template, entry.template_len, 1, fp)) {
 			log_err("Unable to read template\n");
-			return -1;
+			goto out;
 		}
 
 		if (validate)
@@ -1540,8 +1341,6 @@ static int ima_measurement(const char *file)
 			ima_ng_show(&entry);
 	}
 
-	fclose(fp);
-
 	tpm_pcr_read(10, pcr10, sizeof(pcr10));
 
 	log_info("PCRAgg: ");
@@ -1552,10 +1351,14 @@ static int ima_measurement(const char *file)
 
 	if (memcmp(pcr, pcr10, sizeof(pcr))) {
 		log_err("PCRAgg does not match PCR-10\n");
-		return -1;
+		goto out;
 	}
 
-	return 0;
+	err = 0;
+out:
+	fclose(fp);
+
+	return err;
 }
 
 static int cmd_ima_measurement(struct command *cmd)
@@ -1642,36 +1445,39 @@ static void usage(void)
 	printf(
 		"\n"
 		"  -a, --hashalgo     sha1 (default), sha224, sha256, sha384, sha512\n"
-		"  -s, --imasig       also make IMA signature\n"
-		"  -d, --imahash      also make IMA hash\n"
+		"  -s, --imasig       make IMA signature\n"
+		"  -d, --imahash      make IMA hash\n"
 		"  -f, --sigfile      store IMA signature in .sig file instead of xattr\n"
-		"  -1, --rsa          signing key is in RSA DER format (signing v1)\n"
-		"  -k, --key          path to signing key (default keys are /etc/keys/{privkey,pubkey}_evm.pem)\n"
+		"      --rsa          use RSA key type and signing scheme v1\n"
+		"  -k, --key          path to signing key (default: /etc/keys/{privkey,pubkey}_evm.pem)\n"
 		"  -p, --pass         password for encrypted signing key\n"
-		"  -u, --uuid         use file system UUID in HMAC calculation (EVM v2)\n"
-		"  -t, --type         file types to fix 'fdsxm' (f - file, d - directory, s - block/char/symlink)\n"
-		"                     x - skip fixing if both ima and evm xattrs exist (caution: they may be wrong)\n"
+		"  -r, --recursive    recurse into directories (sign)\n"
+		"  -t, --type         file types to fix 'fdsxm' (f: file, d: directory, s: block/char/symlink)\n"
+		"                     x - skip fixing if both ima and evm xattrs exist (use with caution)\n"
 		"                     m - stay on the same filesystem (like 'find -xdev')\n"
 		"  -n                 print result to stdout instead of setting xattr\n"
-		"  -r, --recursive    recurse into directories (sign)\n"
-		"  --m32              force signature for 32 bit target system\n"
-		"  --m64              force signature for 32 bit target system\n"
+		"  -u, --uuid         use custom FS UUID for EVM (unspecified: from FS, empty: do not use)\n"
+		"      --smack        use extra SMACK xattrs for EVM\n"
+		"      --m32          force EVM hmac/signature for 32 bit target system\n"
+		"      --m64          force EVM hmac/signature for 64 bit target system\n"
 		"  -v                 increase verbosity level\n"
 		"  -h, --help         display this help and exit\n"
 		"\n");
 }
 
 struct command cmds[] = {
+	{"--version", NULL, 0, ""},
 	{"help", cmd_help, 0, "<command>"},
 	{"import", cmd_import, 0, "[--rsa] pubkey keyring", "Import public key into the keyring.\n"},
-	{"sign", cmd_sign_evm, 0, "[-r] [--imahash | --imasig ] [--key key] [--pass password] file", "Sign file metadata.\n"},
+	{"sign", cmd_sign_evm, 0, "[-r] [--imahash | --imasig ] [--key key] [--pass [password] file", "Sign file metadata.\n"},
 	{"verify", cmd_verify_evm, 0, "file", "Verify EVM signature (for debugging).\n"},
-	{"ima_sign", cmd_sign_ima, 0, "[--sigfile] [--key key] [--pass password] file", "Make file content signature.\n"},
+	{"ima_sign", cmd_sign_ima, 0, "[--sigfile] [--key key] [--pass [password] file", "Make file content signature.\n"},
 	{"ima_verify", cmd_verify_ima, 0, "file", "Verify IMA signature (for debugging).\n"},
 	{"ima_hash", cmd_hash_ima, 0, "file", "Make file content hash.\n"},
 	{"ima_measurement", cmd_ima_measurement, 0, "file", "Verify measurement list (experimental).\n"},
 	{"ima_fix", cmd_ima_fix, 0, "[-t fdsxm] path", "Recursively fix IMA/EVM xattrs in fix mode.\n"},
-	{"sign_hash", cmd_sign_hash, 0, "[--key key] [--pass password]", "Sign hashes from shaXsum output.\n"},
+	{"ima_clear", cmd_ima_clear, 0, "[-t fdsxm] path", "Recursively remove IMA/EVM xattrs.\n"},
+	{"sign_hash", cmd_sign_hash, 0, "[--key key] [--pass [password]", "Sign hashes from shaXsum output.\n"},
 #ifdef DEBUG
 	{"hmac", cmd_hmac_evm, 0, "[--imahash | --imasig ] file", "Sign file metadata with HMAC using symmetric key (for testing purpose).\n"},
 #endif
@@ -1683,7 +1489,7 @@ static struct option opts[] = {
 	{"imasig", 0, 0, 's'},
 	{"imahash", 0, 0, 'd'},
 	{"hashalgo", 1, 0, 'a'},
-	{"pass", 1, 0, 'p'},
+	{"pass", 2, 0, 'p'},
 	{"sigfile", 0, 0, 'f'},
 	{"uuid", 2, 0, 'u'},
 	{"rsa", 0, 0, '1'},
@@ -1692,10 +1498,46 @@ static struct option opts[] = {
 	{"recursive", 0, 0, 'r'},
 	{"m32", 0, 0, '3'},
 	{"m64", 0, 0, '6'},
+	{"smack", 0, 0, 256},
+	{"version", 0, 0, 257},
 	{}
 
 };
 
+static char *get_password(void)
+{
+	struct termios flags, tmp_flags;
+	char *password, *pwd;
+	int passlen = 64;
+
+	password = malloc(passlen);
+	if (!password) {
+		perror("malloc");
+		return NULL;
+	}
+
+	tcgetattr(fileno(stdin), &flags);
+	tmp_flags = flags;
+	tmp_flags.c_lflag &= ~ECHO;
+	tmp_flags.c_lflag |= ECHONL;
+
+	if (tcsetattr(fileno(stdin), TCSANOW, &tmp_flags) != 0) {
+		perror("tcsetattr");
+		return NULL;
+	}
+
+	printf("PEM password: ");
+	pwd = fgets(password, passlen, stdin);
+
+	/* restore terminal */
+	if (tcsetattr(fileno(stdin), TCSANOW, &flags) != 0) {
+		perror("tcsetattr");
+		return NULL;
+	}
+
+	return pwd;
+}
+
 int main(int argc, char *argv[])
 {
 	int err = 0, c, lind;
@@ -1704,7 +1546,7 @@ int main(int argc, char *argv[])
 	g_argc = argc;
 
 	while (1) {
-		c = getopt_long(argc, argv, "hvnsda:p:fu::k:t:r", opts, &lind);
+		c = getopt_long(argc, argv, "hvnsda:p::fu::k:t:ri", opts, &lind);
 		if (c == -1)
 			break;
 
@@ -1731,21 +1573,31 @@ int main(int argc, char *argv[])
 			params.hash_algo = optarg;
 			break;
 		case 'p':
-			keypass = optarg;
+			if (optarg)
+				params.keypass = optarg;
+			else
+				params.keypass = get_password();
 			break;
 		case 'f':
 			sigfile = 1;
 			xattr = 0;
 			break;
 		case 'u':
-			uuid_str = optarg ?: "+";
+			uuid_str = optarg;
+			if (uuid_str)
+				hmac_flags |= HMAC_FLAG_UUID_SET;
+			else
+				hmac_flags &= ~HMAC_FLAG_UUID;
 			break;
 		case '1':
-			x509 = 0;
+			params.x509 = 0;
 			break;
 		case 'k':
 			params.keyfile = optarg;
 			break;
+		case 'i':
+			evm_immutable = true;
+			break;
 		case 't':
 			search_type = optarg;
 			break;
@@ -1758,6 +1610,13 @@ int main(int argc, char *argv[])
 		case '6':
 			msize = 64;
 			break;
+		case 256:
+			evm_config_xattrnames = evm_extra_smack_xattrs;
+			break;
+		case 257:
+			printf("evmctl %s\n", VERSION);
+			exit(0);
+			break;
 		case '?':
 			exit(1);
 			break;
@@ -1766,14 +1625,6 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	if (x509)
-		sign_hash = sign_hash_v2;
-	else
-		sign_hash = sign_hash_v1;
-
-	OpenSSL_add_all_algorithms();
-	ERR_load_crypto_strings();
-
 	if (argv[optind] == NULL)
 		usage();
 	else
@@ -1781,6 +1632,7 @@ int main(int argc, char *argv[])
 
 	if (err) {
 		unsigned long error;
+
 		if (errno)
 			log_err("errno: %s (%d)\n", strerror(errno), errno);
 		for (;;) {
@@ -1793,6 +1645,6 @@ int main(int argc, char *argv[])
 
 	ERR_free_strings();
 	EVP_cleanup();
-
+	BIO_free(NULL);
 	return err;
 }
